@@ -33,6 +33,7 @@ import shutil
 import threading
 import atexit
 import tempfile
+import pathlib
 try:
   from string import maketrans
 except ImportError:
@@ -434,7 +435,12 @@ class SoftwareProperties(object):
     """Store a backup of the source.list in memory"""
     self.sourceslist_backup = []
     for source in self.sourceslist.list:
-        source_bkp = SourceEntry(line=source.line,file=source.file)
+        if isinstance(source, SourceEntry):
+            source_bkp = SourceEntry(line=source.line,file=source.file)
+        elif isinstance(source, Deb822SourceEntry):
+            source_bkp = Deb822SourceEntry(section=source.line, file=source.file)
+        else:
+            raise TypeError(f"Unknown source: {source}")
         self.sourceslist_backup.append(source_bkp)
 
   def _find_source_from_string(self, line):
@@ -479,7 +485,12 @@ class SoftwareProperties(object):
     """Restore the source list from the startup of the dialog"""
     self.sourceslist.list = []
     for source in self.sourceslist_backup:
-        source_reset = SourceEntry(line=source.line,file=source.file)
+        if isinstance(source, SourceEntry):
+            source_reset = SourceEntry(line=source.line,file=source.file)
+        elif isinstance(source, Deb822SourceEntry):
+            source_reset = Deb822SourceEntry(section=source.line, file=source.file)
+        else:
+            raise TypeError(f"Unknown source: {source}")
         self.sourceslist.list.append(source_reset)
     self.save_sourceslist()
     self.reload_sourceslist()
@@ -500,8 +511,11 @@ class SoftwareProperties(object):
     """Render a nice output to show the source in a treeview"""
     if source.template == None:
         if source.comment:
-            contents = "<b>%s</b> %s" % (escape(source.comment).strip(),
-                                         source.dist)
+            comment = (
+                source.comment.replace("\n", " ").strip()[:80]
+                + ("[…]" if len(source.comment) > 80 else "")
+            )
+            contents = "<b>%s</b> %s" % (escape(comment), source.dist)
             # Only show the components if there are more than one
             if len(source.comps) > 1:
                 for c in source.comps:
@@ -605,6 +619,7 @@ class SoftwareProperties(object):
     self.sourceslist.refresh()
     self.sourceslist_visible=[]
     self.distro.get_sources(self.sourceslist)
+    self.get_source_code_state()
 
   def write_config(self):
     """Write the current apt configuration to file"""
@@ -619,6 +634,9 @@ class SoftwareProperties(object):
         break
     else:
       print("No config found, creating one")
+      dirname = os.path.dirname(f)
+      if not os.path.exists(dirname):
+        pathlib.Path(dirname).mkdir(parents=True)
       with open(conffiles[0], "w") as f:
         f.write("")
 
@@ -785,7 +803,10 @@ class SoftwareProperties(object):
     # file are gone and the file is not saved. we work around that
     # here
     if source.file != apt_pkg.config.find_file("Dir::Etc::sourcelist"):
-      self.sourceslist.list.append(SourceEntry("", file=source.file))
+      if source.file.endswith(".sources"):
+        self.sourceslist.list.append(Deb822SourceEntry("", file=source.file))
+      else:
+        self.sourceslist.list.append(SourceEntry("", file=source.file))
     try:
       self.sourceslist.remove(source)
     except ValueError:
@@ -821,14 +842,6 @@ class SoftwareProperties(object):
     except:
         return False
 
-  def update_keys(self):
-    """ Run apt-key update """
-    try:
-      self.apt_key.update()
-      return True
-    except:
-      return False
-
   def get_package_id(self, apt_cache, pkg):
     """ Return the PackageKit package id """
     assert isinstance(pkg, apt_pkg.Package)
